<html>
<head>
    <title>Example Strava Login Results</title>
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
</head>
<body>
<h1>Authorization Results</h1>
<p>
	Hi, <strong>{{ athlete.firstname }}</strong>.

  Your access token:
  <pre>{{ access_token|tojson(indent=2) }}</pre>
</p>
</body>
</html>
