<html>
<head>
    <title>Example Strava Login Error</title>
    <link rel="stylesheet" href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css" integrity="sha384-BVYiiSIFeK1dGmJRAkycuHAHRg32OmUcww7on3RYdg4Va+PmSTsz/K68vbdEjh4u" crossorigin="anonymous">
</head>
<body>
<h1>Authorization Failed</h1>
<p>
	There was an error authorizing the application to
	access your Strava account.
</p>
<p>
	Error code: <strong>{{ error }}</strong>
</p>

</body>
</html>
